"""
Implementation of models concept on top of Google cloud firestore.
Firestore entities try to make interaction with firestore as simple as possible for the developer.
"""

from firestore.entity import Entity
from firestore.db import SERVER_TIMESTAMP


from google.cloud.firestore import Query as FSQuery
from firestore.db import MalformedQueryError


class Query(object):
    """
    A  query object is returned when you call `firestore.db.Entity.query()`.
    You can iterate over the query to get the results of your query one by one. Each item is an instance of
    `firestore.db.Entity`
    """

    def __init__(self, entity, offset: int, limit: int):
        """Initialize a query

        Args:
            entity: The entity you want to query
            offset: The position to begin the query results
            limit: Maximum number of results to return
        """
        from firestore.entity import __get_client__
        self.__query = __get_client__().collection(entity.__collection_name__())
        if offset:
            self.__query = self.__query.offset(offset)
        if limit:
            self.__query = self.__query.limit(limit)
        self.__fetched = False
        self.__entity = entity
        self.__array_contains_queries = 0
        self.__range_filter_queries = {}

    def __validate_value(self, property_name, value):
        return self.__entity.__dict__.get(property_name).validate(value)

    def __add_range_filter(self, prop):
        self.__range_filter_queries[prop] = True

        # Range filter queries are only allowed on a single property at any given time
        if len(self.__range_filter_queries.keys()) > 1:
            raise MalformedQueryError("Range filter queries i.e (<), (>), (<=) and (>=) "
                                      "can only be performed on a single property in a query")

    def equal(self, prop, value):
        """
        A query condition where prop == value

        Args:
             prop (str): The name of a property to compare
             value (Any): The value to compare from the property

        Returns:
            Query: A query object with this condition added
        """
        self.__query = self.__query.where(prop, "==", self.__validate_value(prop, value))
        return self

    def greater_than(self, prop, value):
        """
        A query condition where prop > value

        Args:
             prop (str): The name of a property to compare
             value (Any): The value to compare from the property

        Returns:
            Query: A query object with this condition added
        """
        self.__add_range_filter(prop)
        self.__query = self.__query.where(prop, ">", self.__validate_value(prop, value))

    def less_than(self, prop, value):
        """
        A query condition where prop < value

        Args:
             prop (str): The name of a property to compare
             value (Any): The value to compare from the property

        Returns:
            Query: A query object with this condition added
        """
        self.__add_range_filter(prop)
        self.__query = self.__query.where(prop, "<", self.__validate_value(prop, value))

    def greater_than_or_equal(self, prop, value):
        """
        A query condition where prop >= value

        Args:
             prop (str): The name of a property to compare
             value (Any): The value to compare from the property

        Returns:
            Query: A query object with this condition added
        """
        self.__add_range_filter(prop)
        self.__query = self.__query.where(prop, ">=", self.__validate_value(prop, value))

    def less_than_or_equal(self, prop, value):
        """
        A query condition where prop <= value

        Args:
             prop (str): The name of a property to compare
             value (Any): The value to compare from the property

        Returns:
            Query: A query object with this condition added
        """
        self.__add_range_filter(prop)
        self.__query = self.__query.where(prop, "<=", self.__validate_value(prop, value))

    def contains(self, prop, value):
        """
        A query condition where `value in prop`

        Args:
             prop (str): The name of a property to compare
             value (Any): The value to compare from the property

        Returns:
            Query: A query object with this condition added

        Raises:
            MalformedQueryError: If the property specified is not a ListField, or
               the query has more than one contains condition
        """

        # Make sure there's only on `array_contains` condition
        self.__array_contains_queries += 1
        if self.__array_contains_queries > 1:
            raise MalformedQueryError("Only one `contains` clause is allowed per query")
        self.__query = self.__query.where(prop, "array_contains", value)
        return self

    def order_by(self, prop, direction="ASC"):
        """
        Set an order for the query, accepts

        Args:
            prop (str): The property name to order by
            direction (str: "ASC" or "DESC"), optional:

        Returns:
             Query: A query object with order applied
        """
        if direction != "ASC" and direction != "DESC":
            raise MalformedQueryError("order_by direction can only be ASC, or DESC")
        direction = FSQuery.ASCENDING if direction == "ASC" else FSQuery.DESCENDING
        self.__query = self.__query.order_by(prop, direction=direction)
        return self

    def __fetch(self):
        self.__docs = self.__query.get()
        self.__fetched = True

    def __iter__(self):
        return self

    def fetch(self):
        """
        Get the results of the query as a list

        Returns:
            list (`firestore.db.Entity`): A list of entities for the found results
        """
        return [entity for entity in self]

    def __next__(self):
        # Fetch data from db if not already done
        if not self.__fetched:
            self.__fetch()
        doc = self.__docs.__next__()
        entity = self.__entity(id=doc.id)
        entity.__firestore_data__ = doc.to_dict()
        return entity



"""
This is a collection of tools used by mantle Database packages, the include property types and common errors
"""

import json
import pickle
from datetime import datetime, date
from google.cloud.firestore import SERVER_TIMESTAMP


class Property(object):
    """
    A class describing a typed, persisted attribute of a database entity
    """
    def __init__(self, default=None, required=False, repeated=False):
        """
        Args:
            default: The default value of the property
            required: Enforce the property value to be provided
            repeated (bool): Stores multiple values as a list, Overrides default with []
        """
        if type(self) is Property:
            raise Exception("You must extend Property")
        if repeated:
            if default is not None:
                raise Exception("`default` is not allowed for a repeated property.")
            default = []
        self.default = default
        self.required = required
        self.name = None
        self.repeated = repeated

    def __set_name__(self, owner, name):
        self.name = name

    def __set__(self, instance, value):
        if value is None:
            instance.__firestore_data__[self.name] = None
        if self.repeated:
            value = self.__type_check__(value, list)
            if value:
                value = [self.validate(val) for val in value]
        else:
            value = self.validate(value)
        instance.__firestore_data__[self.name] = value

    def __get__(self, instance, owner):
        value = instance.__firestore_data__.get(self.name)
        if value:
            if self.repeated:
                value = [self.user_value(val) for val in value]
            else:
                value = self.user_value(value)
        elif (value is None) and self.repeated:
            return []
        return value

    def __type_check__(self, user_value, data_types):
        """
        Check whether this value has the right data type
        Args:
            user_value: The user_value you want to confirm
            data_types: Type/Types to check against

        Returns:
            user_value: A type checked user value or the default value
        """
        if self.required and self.default is None and user_value is None:
            raise InvalidValueError(self, user_value)
        # Assign a default value if None is provided
        if user_value is None:
            user_value = self.default

        if not isinstance(user_value, data_types) and user_value is not None:
            raise InvalidValueError(self, user_value)
        return user_value

    def validate(self, value):
        raise NotImplementedError("A property must implement validate")

    def user_value(self, value):
        return value


class TextProperty(Property):
    """An Property whose value is a text string of unlimited length.
    I'ts not advisable to index this property
    """
    def validate(self, value):
        return self.__type_check__(value, str)


class StringProperty(Property):
    """
    An indexed Property whose value is a text string of limited length.

    Args:
        default: Default value for this property
        length (int=255): The maximum length of this property
        required (bool): Enforce whether this value can be empty
        repeated (bool): Stores multiple values as a list, Overrides default with []
    """
    def __init__(self, default=None, length=255, required=False, repeated=False):
        super(StringProperty, self).__init__(default=default, required=required, repeated=repeated)
        self.length = length

    def validate(self, value):
        value = self.__type_check__(value, str)
        if value is not None and len(value) > self.length:
            raise InvalidValueError(self, value)
        return value


class IntegerProperty(Property):
    """A Property whose value is a Python int or long"""
    def validate(self, value):
        return self.__type_check__(value, int)


class FloatingPointNumberProperty(Property):
    """A Property whose value is a Python float.

    Note: int and long are also allowed.
    """
    def validate(self, value):
        return self.__type_check__(value, (int, float))


class BlobProperty(Property):
    """A Property whose value is a byte string. It may be compressed."""
    def validate(self, value):
        return self.__type_check__(value, bytes)


class ReferenceProperty(Property):
    def __init__(self, entity, required=False, repeated=False):
        from firestore import Entity
        if not issubclass(entity, Entity):
            raise ReferencePropertyError("A reference property must reference another entity")
        super(ReferenceProperty, self).__init__(required=required, repeated=repeated)
        self.entity = entity

    def user_value(self, document):
        """
        Returns:
            Entity: The value of the field
        """
        if not document:
            return None
        entity = self.entity(id=document.id)
        entity.__firestore_data__ = document.get().to_dict()
        return entity

    def validate(self, value):
        value = self.__type_check__(value, self.entity)
        if not value:
            return None
        if not value.id:
            raise ReferencePropertyError("A reference must be put first before it can be referenced")
        return value.__document__()


class DictProperty(Property):
    """A property whose value is any Json-encodable Python object.
    """
    def __init__(self, required=False, repeated=False):
        super(DictProperty, self).__init__(required=required, default={}, repeated=repeated)

    def validate(self, value):
        if isinstance(value, str):
            try:
                value = json.loads(value)
            except json.JSONDecodeError:
                raise InvalidPropertyError(self, "JsonProperty must be valid JSON")
        return self.__type_check__(value, dict)


class BooleanProperty(Property):
    """A Property whose value is a Python bool."""
    def validate(self, value):
        return self.__type_check__(value, bool)


class DateTimeProperty(Property):
    """A Property whose value is a datetime object.

    Note: auto_now_add can be overridden by setting the value before writing the entity.
    Args:
        default (datetime)
        required (bool): Enforce that this property can't be submitted when empty
        auto_now (bool): Set to the current time every time the model is updated
        auto_add_now (bool): Set to the current time when a record is created
    """
    def __init__(self, default=None, required=False, auto_now=False, auto_add_now=False, repeated=False):
        if not default and auto_add_now:
            default = SERVER_TIMESTAMP
        super(DateTimeProperty, self).__init__(default=default, required=required, repeated=repeated)
        self.auto_now = auto_now

    def user_value(self, value):
        """
        Returns:
            datetime: The value of the field
        """
        if value == SERVER_TIMESTAMP:
            return datetime.now()
        return value

    def validate(self, value):
        # Return server timestamp as the value
        if value == SERVER_TIMESTAMP or self.auto_now:
            return SERVER_TIMESTAMP
        if value is None and self.default == SERVER_TIMESTAMP:
            return SERVER_TIMESTAMP
        return self.__type_check__(value, datetime)


class DateProperty(Property):
    """A Property whose value is a date object.

    Args:
        default (datetime): The default value for this property
        required (bool): Enforce that this property can't be submitted when empty
        auto_add_now (bool): Set to the current date when a record is created
    """
    def __init__(self, default=None, required=False, auto_add_now=False, repeated=False):
        """
        Returns:
            date: The value of the field
        """
        if not default and auto_add_now:
            default = SERVER_TIMESTAMP
        super(DateProperty, self).__init__(default=default, required=required, repeated=repeated)

    def user_value(self, value):
        if value == SERVER_TIMESTAMP:
            value = datetime.now().date()
        if isinstance(value, datetime):
            return value.date()
        return value

    def validate(self, value):
        # Return server timestamp as the value
        if value is None and self.default == SERVER_TIMESTAMP:
            return SERVER_TIMESTAMP
        if isinstance(value, datetime):
            return value.date()
        return self.__type_check__(value, date)


class PickledProperty(Property):
    """A Property whose value is any picklable Python object."""
    def __int__(self, required=False, repeated=False):
        super(PickledProperty, self).__init__(default=None, required=required, repeated=repeated)

    def user_value(self, value):
        if not value:
            return None
        return pickle.loads(value)

    def validate(self, value):
        if value:
            return pickle.dumps(value)


class InvalidValueError(ValueError):
    """Raised if the value of a property does not fit the property type"""

    def __init__(self, _property, value):
        self.property = _property
        self.value = value

    def __str__(self):
        return "%s is not a valid value for property %s of type %s" % \
               (self.value, self.property.name, type(self.property).__name__)


class MalformedQueryError(Exception):
    """Raised when the rules of a query are broken"""

    def __init__(self, message):
        self.message = message

    def __str__(self):
        return self.message


class InvalidPropertyError(Exception):
    """Raised if a non-existent property is provided during the creation of a model"""

    def __init__(self, prop_name, model_name):
        self.prop_name = prop_name
        self.model_name = model_name

    def __str__(self):
        return "%s not found in model %s" % (self.prop_name, self.model_name)


class ReferencePropertyError(Exception):
    """Raised when a reference property point's to a location the model can't resolve"""
    def __init__(self, message):
        self.message = message

    def __str__(self):
        return self.message



from google.cloud.firestore import Client, SERVER_TIMESTAMP
from firestore.db import Property
from firestore.query import Query
import os


def __get_client__():
    # There must be a better way to do this, right?
    import builtins
    if not hasattr(builtins, "__firestore_client__"):
        setattr(builtins, "__firestore_client__", Client())
    return builtins.__firestore_client__


class Entity(object):
    """Creates a firestore document under the collection [YourEntity]

    Args:
        **data (kwargs): Values for properties in the new record, e.g User(name="Bob")

    Attributes:
        id (str or int): Unique id identifying this record,
            if auto-generated, this is not available before `put()`
    """

    def __init__(self, **data):
        if type(self) is Entity:
            raise Exception("You must extend Entity")
        self.__model_name = type(self).__name__
        client = __get_client__()
        self.__collection = client.collection(self.__collection_name__())
        self.id = None
        self.__firestore_data__ = {}
        if "id" in data:
            self.id = data.pop("id")
        for key, value in data.items():
            setattr(self, key, value)

    @classmethod
    def __collection_name__(cls):
        prefix = os.environ.get('TEST_PREFIX', '')
        return "%s%s" % (prefix, cls.__name__)

    def __document__(self):
        if not self.id:
            return None
        # Get's the absolute path: `projects/{project_id}/databases/{database_id}/documents/{document_path}
        return self.__collection.document(self.id)

    def __str__(self):
        if self.id:
            return "<%s id='%s'>" % (self.__model_name, self.id)
        return "<Entity %s>" % self.__model_name

    def put(self):
        """
        Save the models data to Firestore

        Raises:
            InvalidValueError: Raised if the value of a property is invalid, e.g. A required property that's None
        """
        props = self.__get_properties__()
        for key in props:
            if key not in self.__firestore_data__:
                # Value not set yet. Try to set to none
                setattr(self, key, None)

        if self.id:
            self.__collection.document(self.id).set(self.__firestore_data__)
        else:
            _time, new_ref = self.__collection.add(self.__firestore_data__)
            self.id = new_ref.id
        # Set the data to the most recent data on db. Updates server set values e.g. SERVER_TIMESTAMP
        self.__firestore_data__ = self.__document__().get().to_dict()

    def delete(self):
        """Delete the document connected to this model from firestore"""
        if self.id:
            self.__collection.document(self.id).delete()

    def __get_properties__(self):
        props = []
        for key, value in self.__class__.__dict__.items():
            try:
                if issubclass(value.__class__, Property):
                    props.append(key)
            except TypeError:
                pass
        return props

    def __set_database_values__(self, document: dict):
        self.__firestore_data__ = document

    @classmethod
    def get(cls, _id):
        """
        Get a model with the given id

        Args:
            _id (str or int): A key or id of the model record, when a list is provided, `get` returns a list
                models

        Returns:
            Entity: An instance of the firestore entity calling get
            None: If the id provided doesn't exist
        """
        document = __get_client__().collection(cls.__collection_name__()).document(_id)
        doc = document.get()
        if not doc.exists:
            return None
        entity = cls(id=_id)
        entity.__firestore_data__ = doc.to_dict()
        return entity

    @classmethod
    def query(cls, offset=0, limit=0):
        """
        Create a query to this model

        Args:
            offset (int): The position in the database where the results begin
            limit (int): Maximum number of records to return

        Returns:
            An iterable query object
        """
        return Query(cls, offset, limit)
